package ca.spottedleaf.moonrise.common.util;

import ca.spottedleaf.concurrentutil.executor.thread.PrioritisedThreadPool;
import ca.spottedleaf.moonrise.common.PlatformHooks;
import com.mojang.logging.LogUtils;
import io.canvasmc.canvas.ChunkSystemAlgorithm;
import io.canvasmc.canvas.Config;
import io.canvasmc.canvas.server.chunk.TheChunkSystem;
import io.canvasmc.canvas.util.ThreadBuilder;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

public final class MoonriseCommon {

    private static final Logger LOGGER = LogUtils.getClassLogger();

    // Canvas start - rewrite moonrise executor
    public static TheChunkSystem WORKER_POOL;
    public static TheChunkSystem.ExecutorGroup PARALLEL_GEN_GROUP;
    public static TheChunkSystem.ExecutorGroup RADIUS_AWARE_GROUP;
    public static TheChunkSystem.ExecutorGroup LOAD_GROUP;

    public static void init(final int configWorkerThreads, final int configIoThreads) {
        ChunkSystemAlgorithm algorithm = Config.INSTANCE.chunks.defaultThreadAlgorithm;
        int workerThreads = algorithm.evalWorkers(configWorkerThreads, configIoThreads);
        int ioThreads = algorithm.evalIO(configWorkerThreads, configIoThreads);

        WORKER_POOL = buildChunkSystem(workerThreads);

        PARALLEL_GEN_GROUP = MoonriseCommon.WORKER_POOL.createExecutorGroup();
        RADIUS_AWARE_GROUP = MoonriseCommon.WORKER_POOL.createExecutorGroup();
        LOAD_GROUP = MoonriseCommon.WORKER_POOL.createExecutorGroup();

        IO_POOL.adjustThreadCount(ioThreads);
        LOGGER.info("Running LS ChunkSystem with {} worker threads and {} io threads", workerThreads, ioThreads);
        if ((workerThreads + ioThreads + Config.INSTANCE.ticking.allocatedSchedulerThreadCount) > Runtime.getRuntime().availableProcessors()) {
            LOGGER.warn("Your configuration is not optimal for Canvas to use your CPU efficiently. It should meet this requirement: '(worker count + io count + scheduler count) < max threads'");
            LOGGER.warn("This is due to Canvas' chunk system being much more aggressive compared to other server forks, leading to not meeting this requirement causing major performance issues.");
        }
    }

    private static @NotNull TheChunkSystem buildChunkSystem(int workerThreads) {
        return new TheChunkSystem(workerThreads, new ThreadBuilder() {
            @Override
            public void accept(final Thread thread) {
                thread.setPriority(Config.INSTANCE.chunks.threadPoolPriority);
                thread.setDaemon(true);
                thread.setUncaughtExceptionHandler((thread1, throwable) -> LOGGER.error("Uncaught exception in thread {}", thread1.getName(), throwable));
                thread.setName("ls_wg worker #" + getAndIncrementId());
            }
        }, "ls_wg");
    }
    // Canvas end

    public static final PrioritisedThreadPool IO_POOL = new PrioritisedThreadPool(
        new Consumer<>() {
            private final AtomicInteger idGenerator = new AtomicInteger();
            @Override
            public void accept(final Thread thread) {
                thread.setDaemon(true);
                thread.setName(PlatformHooks.get().getBrand() + " I/O Worker #" + this.idGenerator.getAndIncrement());
                thread.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
                    @Override
                    public void uncaughtException(final Thread thread, final Throwable throwable) {
                        LOGGER.error("Uncaught exception in thread " + thread.getName(), throwable);
                    }
                });
            }
        }
    );
    public static final long IO_QUEUE_HOLD_TIME = (long)(100.0e6); // 100ms
    // Canvas
    public static final PrioritisedThreadPool.ExecutorGroup SERVER_REGION_IO_GROUP = IO_POOL.createExecutorGroup(1, 0);

    public static void haltExecutors() {
        // Canvas start
        LOGGER.info("Terminating ChunkSystem...");
        MoonriseCommon.WORKER_POOL.shutdown();
        // Canvas end
        MoonriseCommon.IO_POOL.shutdown(false);
        LOGGER.info("Awaiting termination of I/O pool for up to 60s...");
        if (!MoonriseCommon.IO_POOL.join(TimeUnit.SECONDS.toMillis(60L))) {
            LOGGER.error("I/O pool did not shut down in time!");
            MoonriseCommon.IO_POOL.halt(false);
        }
    }

    private MoonriseCommon() {}
}
